(ns document-storage.migrate
  (:require [document-storage.core :as ds]
            [document-storage.postgres.core :as dspostgres]
            [environ.core :refer [env]]
            [clojure.tools.logging :as log]
            [editscript.core :as escript]))


(defn migrate-repository [repo-id ds-source ds-target]
  ;;;TODO git is slow, load all git document versions in parallel
  (let [documents (map-indexed vector (.list-documents ds-source repo-id))
        doc-count (count documents)]
    (doall
     (pmap (fn [[i doc-id]]
             (let [dinfo (str "[" repo-id "][" i  "/" doc-count "]")]
               (log/info  dinfo "Start Migrating " doc-id)
      ;;;TODO check if repository is versioned
               (let [target-doc (.load-document ds-target doc-id repo-id)
                     source-doc (.load-document ds-source doc-id repo-id)]
                 (if (and target-doc
                          (= target-doc
                             source-doc))
                   (log/info dinfo "Document already migrated " doc-id)
                   (if (.versioned-repository? ds-source repo-id)
                     (let [versions (sort-by (comp #(. % getTime) :date) < (.list-versions ds-source doc-id repo-id))]
                       (doseq [dv versions]
                         (if (= (:type dv) :delete)
                           (do
                             (log/info dinfo "Document " doc-id " - Version: " (:version dv) " - Type: DELETE")
                             (.delete-document ds-target doc-id repo-id))
                           (let [source-document (if (= (count versions) 1)
                                                   source-doc
                                                   (.load-version ds-source doc-id (:version dv) repo-id))]
                             (log/info dinfo "Save Document " doc-id " - Version: " (:version dv))
                             (.save-document ds-target doc-id source-document repo-id #_{:force-keyframe true})
                             (if (= source-document (.load-document ds-target doc-id repo-id))
                               (log/info dinfo "Check Document " doc-id " - Version: " (:version dv) " passed")
                               (log/error dinfo "Check Document " doc-id " - Version: " (:version dv) " failed"))))))
                     (do
                       (.save-document ds-target doc-id source-doc repo-id)
                       (if (= source-doc (.load-document ds-target doc-id repo-id))
                         (log/info dinfo "Check Document " doc-id " passed")
                         (log/error dinfo "Check Document " doc-id " failed"))))))
               (log/info dinfo "Finished Migrating " doc-id))) documents))))

(defn migrate [ds-source ds-destination]
  (let [repositories (.list-repositories ds-source)]
    (doseq [r repositories]
      (log/info "Migrate Repository " (:id r))
      (migrate-repository (:id r) ds-source ds-destination)
      (log/info "Finished Migrating " (:id r)))))


#_(def gitstorage (ds/make-git-ds "doc-storage"))
#_(def pgstorage (dspostgres/make-postgres-ds
                  {:id :mtesting
                   :repositories {:atoms {:caching true}
                                  :courselists {:caching true}
                                  :courses {:caching true}
                                  :default {:caching true}
                                  :exams {:caching true}
                                  :exam-sessions {}
                                  :linklists {:versioning false}
                                  :live-surveys {:versioning false}
                                  :quiz-sessions {}
                                  :quizzes {:caching true}
                                  :test {:caching true}
                                  :units {:caching true}
                                  :unit-sessions {:versioning false}}
                   :jdbc-url (env :database-url)}))

#_(time (migrate gitstorage pgstorage))
;;;TODO add caching for pgstorage
#_(.list-repositories gitstorage)
#_(.list-repositories pgstorage)
#_(document-storage.postgres.documents/get-last-document-version pgstorage "" :atoms)
#_(.versioned-repository? pgstorage :atoms)
#_(.cached-repository? pgstorage :linklists)
#_(time (do (.load-document pgstorage "20210304T190941-uc82822" :units)
            (+ 1 1)))

#_(time (do (.load-document pgstorage "q2018-10-06T11-03-59-G__1589" :quizzes)
            (+ 1 1)))

#_(defn test-version [storage id repository]
    (let [version (first (take-last 2 (.list-versions storage id repository)))]
      (.load-version storage id (:version version) repository)))

#_(time (test-version gitstorage "q2018-10-06T11-03-59-G__1589" :quizzes))
#_(time (test-version pgstorage "q2018-10-06T11-03-59-G__1589" :quizzes))
#_(time (doseq [d ["q2018-04-19T11-26-06-G__64819"
                   "q2020-11-09T08-06-25-G__486"
                   "q2018-04-19T11-26-05-G__64811"
                   "q2018-04-19T11-26-06-G__64817"
                   "q2020-11-09T09-24-36-G__489"
                   "q2018-04-19T11-26-07-G__64820"
                   "q2018-04-19T11-26-07-G__64821"
                   "q2018-04-19T11-26-06-G__64816"
                   "q2018-04-19T11-26-10-G__64823"
                   "q2018-04-19T11-26-04-G__64808"
                   "q2018-04-19T11-26-10-G__64824"
                   "q2020-11-09T08-06-05-G__294"
                   "q2018-04-19T11-26-05-G__64812"
                   "q2020-11-09T08-08-49-G__1000"
                   "q2018-04-19T11-26-09-G__64822"
                   "q2018-04-19T11-26-04-G__64809"
                   "q2018-05-23T20-35-52-G__1527"
                   "q2018-04-19T11-26-06-G__64818"]]
          (.load-document pgstorage d :quizzes)
          (* 1 1)))

#_(time (do (.load-document gitstorage "u2019-01-14T22-08-05-G__417" :units)
            (+ 11 1)))


#_(let [previous-version (.load-version gitstorage "20210927T120641-atom80156" "ca202f62c216956095d483884ff0641b63ee0b91" :atoms)
        new-version (.load-version gitstorage "20210927T120641-atom80156" "001053e57e989041887bddd07672ec9e6cc7d6fb" :atoms)]
    #_(.save-document pgstorage "20210927T120641-atom80156" new-version "atoms")
    #_previous-version
    (escript/diff previous-version new-version))

#_(let [previous-version (.load-version gitstorage "20210927T120641-atom80156" "ca202f62c216956095d483884ff0641b63ee0b91" :atoms)
        previous-pg-version (.load-document pgstorage "20210927T120641-atom80156" :atoms)
        new-version (.load-version gitstorage "20210927T120641-atom80156" "001053e57e989041887bddd07672ec9e6cc7d6fb" :atoms)]
    #_(.save-document pgstorage "20210927T120641-atom80156" new-version "atoms")
    #_previous-version
    #_(= previous-version previous-pg-version)
    ;;;FIXME this will crash with a-star?!
    #_(escript/diff previous-pg-version new-version {:str-diff? true :algo :a-star})

    previous-pg-version)
#_(time (.load-document pgstorage "u2019-01-14T22-08-05-G__417" :units))

#_(time (.load-version pgstorage "u2019-01-14T22-08-05-G__417"  160 :units))



#_(= (.load-version gitstorage "20180926T110658-atom554" "ebc38a8856a2a86a5d6cde415824ce0300fbff64" #_"4357f94291c0e2e2df758f1fd61c1c06f9ce7bbf" :atoms)
     (.load-document gitstorage "20180926T110658-atom554" :atoms))
#_(last (sort-by (comp #(. % getTime) :date) < (.list-versions gitstorage "20180926T110658-atom554" :atoms)))
#_(map (comp println :commit) (sort-by (comp #(. % getTime) :date) < (.list-versions gitstorage "20180926T110658-atom554" :atoms)))

#_(sort-by (comp #(. % getTime) :date) < (.list-versions gitstorage "20180926T110658-atom554" :atoms))


#_(.list-documents gitstorage :atoms)
#_(let [repo (first (.list-repositories gitstorage))
        repo-id (:id repo)
        doc (nth (.list-documents gitstorage repo-id) 16)
        versions (.list-versions gitstorage doc repo-id)
        version (first versions)]
    (.load-version gitstorage doc (:version version) repo-id))

#_((comp #(. % getTime) :time) {:time (new java.util.Date)})
